{******************************************************************************}
{* DCPcrypt v2.0 written by David Barton (crypto@cityinthesky.co.uk) **********}
{******************************************************************************}
{* A binary compatible implementation of Blowfish *****************************}
{******************************************************************************}
{* Copyright (c) 1999-2002 David Barton                                       *}
{* Permission is hereby granted, free of charge, to any person obtaining a    *}
{* copy of this software and associated documentation files (the "Software"), *}
{* to deal in the Software without restriction, including without limitation  *}
{* the rights to use, copy, modify, merge, publish, distribute, sublicense,   *}
{* and/or sell copies of the Software, and to permit persons to whom the      *}
{* Software is furnished to do so, subject to the following conditions:       *}
{*                                                                            *}
{* The above copyright notice and this permission notice shall be included in *}
{* all copies or substantial portions of the Software.                        *}
{*                                                                            *}
{* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR *}
{* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,   *}
{* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    *}
{* THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER *}
{* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING    *}
{* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER        *}
{* DEALINGS IN THE SOFTWARE.                                                  *}
{******************************************************************************}
UNIT blowfish;

INTERFACE
USES
  Classes, Sysutils, crypt2, ciphers;
CONST
  id_blowfish=32;

TYPE
  TDCP_blowfish= CLASS(TDCP_blockcipher64)
  PROTECTED
    SBox: ARRAY[0..3,0..255] OF DWord;
    PBox: ARRAY[0..17] OF DWord;
    PROCEDURE InitKey(CONST Key; Size: longword); override;
  PUBLIC
    CLASS FUNCTION GetId: INTEGER; override;
    CLASS FUNCTION GetAlgorithm: string; override;
    CLASS FUNCTION GetMaxKeySize: INTEGER; override;
    CLASS FUNCTION SelfTest: BOOLEAN; override;
    PROCEDURE Burn; override;
    PROCEDURE EncryptECB(CONST InData; VAR OutData); override;
    PROCEDURE DecryptECB(CONST InData; VAR OutData); override;
  END;


{******************************************************************************}
{******************************************************************************}
IMPLEMENTATION
{$R-}{$Q-}
{$I blowfish.inc}

CLASS FUNCTION TDCP_blowfish.GetID: INTEGER;
BEGIN
  Result:= id_blowfish;
END;

CLASS FUNCTION TDCP_blowfish.GetAlgorithm: string;
BEGIN
  Result:= 'Blowfish';
END;

CLASS FUNCTION TDCP_blowfish.GetMaxKeySize: INTEGER;
BEGIN
  Result:= 448;
END;

CLASS FUNCTION TDCP_blowfish.SelfTest: BOOLEAN;
CONST
  Key1: ARRAY[0..7] OF BYTE= ($00,$00,$00,$00,$00,$00,$00,$00);
  Key2: ARRAY[0..7] OF BYTE= ($7C,$A1,$10,$45,$4A,$1A,$6E,$57);
  InData1: ARRAY[0..7] OF BYTE= ($00,$00,$00,$00,$00,$00,$00,$00);
  InData2: ARRAY[0..7] OF BYTE= ($01,$A1,$D6,$D0,$39,$77,$67,$42);
  OutData1: ARRAY[0..7] OF BYTE= ($4E,$F9,$97,$45,$61,$98,$DD,$78);
  OutData2: ARRAY[0..7] OF BYTE= ($59,$C6,$82,$45,$EB,$05,$28,$2B);
VAR
  Cipher: TDCP_blowfish;
  Data: ARRAY[0..7] OF BYTE;
BEGIN
  Cipher:= TDCP_blowfish.Create;
  Cipher.Init(Key1,Sizeof(Key1)*8,NIL);
  Cipher.EncryptECB(InData1,Data);
  Result:= BOOLEAN(CompareMem(@Data,@OutData1,Sizeof(Data)));
  Cipher.Reset;
  Cipher.DecryptECB(Data,Data);
  Result:= BOOLEAN(CompareMem(@Data,@InData1,Sizeof(Data))) AND Result;
  Cipher.Burn;
  Cipher.Init(Key2,Sizeof(Key2)*8,NIL);
  Cipher.EncryptECB(InData2,Data);
  Result:= BOOLEAN(CompareMem(@Data,@OutData2,Sizeof(Data))) AND Result;
  Cipher.Reset;
  Cipher.DecryptECB(Data,Data);
  Result:= BOOLEAN(CompareMem(@Data,@InData2,Sizeof(Data))) AND Result;
  Cipher.Burn;
  Cipher.Free;
END;

PROCEDURE TDCP_blowfish.InitKey(CONST Key; Size: longword);
VAR
  i, k: longword;
  A: DWord;
  KeyB: PByteArray;
  Block: ARRAY[0..7] OF BYTE;
BEGIN
  Size:= Size div 8;
  KeyB:= @Key;
  Move(SBoxOrg,SBox,Sizeof(SBox));
  Move(PBoxOrg,PBox,Sizeof(PBox));
  k:= 0;
  FOR i:= 0 TO 17 DO
  BEGIN
    A:= dword(KeyB^[(k+3) mod Size]);
    A:= A + (dword(KeyB^[(k+2) mod Size]) shl 8);
    A:= A + (dword(KeyB^[(k+1) mod Size]) shl 16);
    A:= A + (dword(KeyB^[k]) shl 24);
    PBox[i]:= PBox[i] xor A;
    k:= (k+4) mod Size;
  END;
  FillChar(Block,Sizeof(Block),0);
  FOR i:= 0 TO 8 DO
  BEGIN
    EncryptECB(Block,Block);
    PBox[i*2]:= dword(Block[3]) + (dword(Block[2]) shl 8) + (dword(Block[1]) shl 16) + (dword(Block[0]) shl 24);
    PBox[i*2+1]:= dword(Block[7]) + (dword(Block[6]) shl 8) + (dword(Block[5]) shl 16) + (dword(Block[4]) shl 24);
  END;
  FOR k:= 0 TO 3 DO
  BEGIN
    FOR i:= 0 TO 127 DO
    BEGIN
      EncryptECB(Block,Block);
      SBox[k,i*2]:= dword(Block[3]) + (dword(Block[2]) shl 8) + (dword(Block[1]) shl 16) + (dword(Block[0]) shl 24);
      SBox[k,i*2+1]:= dword(Block[7]) + (dword(Block[6]) shl 8) + (dword(Block[5]) shl 16) + (dword(Block[4]) shl 24);
    END;
  END;
END;

PROCEDURE TDCP_blowfish.Burn;
BEGIN
  FillChar(SBox,Sizeof(SBox),$FF);
  FillChar(PBox,Sizeof(PBox),$FF);
  inherited Burn;
END;

PROCEDURE TDCP_blowfish.EncryptECB(CONST InData; VAR OutData);
VAR
  xL, xR: DWord;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  xL:= Pdword(@InData)^;
  xR:= Pdword(longword(@InData)+4)^;
  xL:= ((xL AND $FF) shl 24) OR ((xL AND $FF00) shl 8) OR ((xL AND $FF0000) shr 8) OR ((xL AND $FF000000) shr 24);
  xR:= ((xR AND $FF) shl 24) OR ((xR AND $FF00) shl 8) OR ((xR AND $FF0000) shr 8) OR ((xR AND $FF000000) shr 24);
  xL:= xL xor PBox[0];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[1];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[2];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[3];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[4];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[5];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[6];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[7];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[8];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[9];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[10];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[11];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[12];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[13];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[14];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[15];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[16];
  xR:= xR xor PBox[17];
  xL:= ((xL AND $FF) shl 24) OR ((xL AND $FF00) shl 8) OR ((xL AND $FF0000) shr 8) OR ((xL AND $FF000000) shr 24);
  xR:= ((xR AND $FF) shl 24) OR ((xR AND $FF00) shl 8) OR ((xR AND $FF0000) shr 8) OR ((xR AND $FF000000) shr 24);
  Pdword(@OutData)^:= xR;
  Pdword(longword(@OutData)+4)^:= xL;
END;

PROCEDURE TDCP_blowfish.DecryptECB(CONST InData; VAR OutData);
VAR
  xL, xR: DWord;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  xL:= Pdword(@InData)^;
  xR:= Pdword(longword(@InData)+4)^;
  xL:= (xL shr 24) OR ((xL shr 8) AND $FF00) OR ((xL shl 8) AND $FF0000) OR (xL shl 24);
  xR:= (xR shr 24) OR ((xR shr 8) AND $FF00) OR ((xR shl 8) AND $FF0000) OR (xR shl 24);
  xL:= xL xor PBox[17];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[16];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[15];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[14];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[13];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[12];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[11];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[10];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[9];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[8];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[7];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[6];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[5];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[4];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[3];
  xR:= xR xor (((SBox[0,(xL shr 24) AND $FF] + SBox[1,(xL shr 16) AND $FF]) xor
    SBox[2,(xL shr 8) AND $FF]) + SBox[3,xL AND $FF]) xor PBox[2];
  xL:= xL xor (((SBox[0,(xR shr 24) AND $FF] + SBox[1,(xR shr 16) AND $FF]) xor
    SBox[2,(xR shr 8) AND $FF]) + SBox[3,xR AND $FF]) xor PBox[1];
  xR:= xR xor PBox[0];
  xL:= (xL shr 24) OR ((xL shr 8) AND $FF00) OR ((xL shl 8) AND $FF0000) OR (xL shl 24);
  xR:= (xR shr 24) OR ((xR shr 8) AND $FF00) OR ((xR shl 8) AND $FF0000) OR (xR shl 24);
  Pdword(@OutData)^:= xR;
  Pdword(longword(@OutData)+4)^:= xL;
END;

END.
