{******************************************************************************}
{* DCPcrypt v2.0 written by David Barton (crypto@cityinthesky.co.uk) **********}
{******************************************************************************}
{* Block cipher component definitions *****************************************}
{******************************************************************************}
{* Copyright (c) 1999-2002 David Barton                                       *}
{* Permission is hereby granted, free of charge, to any person obtaining a    *}
{* copy of this software and associated documentation files (the "Software"), *}
{* to deal in the Software without restriction, including without limitation  *}
{* the rights to use, copy, modify, merge, publish, distribute, sublicense,   *}
{* and/or sell copies of the Software, and to permit persons to whom the      *}
{* Software is furnished to do so, subject to the following conditions:       *}
{*                                                                            *}
{* The above copyright notice and this permission notice shall be included in *}
{* all copies or substantial portions of the Software.                        *}
{*                                                                            *}
{* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR *}
{* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,   *}
{* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    *}
{* THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER *}
{* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING    *}
{* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER        *}
{* DEALINGS IN THE SOFTWARE.                                                  *}
{******************************************************************************}
UNIT ciphers;

INTERFACE
USES
  Classes, Sysutils, crypt2;

{******************************************************************************}
    { Base type definition for 64 bit block ciphers }
TYPE
  TDCP_blockcipher64= CLASS(TDCP_blockcipher)
  PRIVATE
    IV, CV: ARRAY[0..7] OF BYTE;

    PROCEDURE IncCounter;
  PUBLIC
    CLASS FUNCTION GetBlockSize: INTEGER; override;
      { Get the block size of the cipher (in bits) }

    PROCEDURE Reset; override;
      { Reset any stored chaining information }
    PROCEDURE Burn; override;
      { Clear all stored key information and chaining information }
    PROCEDURE SetIV(CONST Value); override;
      { Sets the IV to Value and performs a reset }
    PROCEDURE GetIV(VAR Value); override;
      { Returns the current chaining information, not the actual IV }
    PROCEDURE Init(CONST Key; Size: longword; InitVector: POINTER); override;
      { Do key setup based on the data in Key, size is in bits }

    PROCEDURE EncryptCBC(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CBC method of encryption }
    PROCEDURE DecryptCBC(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CBC method of decryption }
    PROCEDURE EncryptCFB8bit(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CFB (8 bit) method of encryption }
    PROCEDURE DecryptCFB8bit(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CFB (8 bit) method of decryption }
    PROCEDURE EncryptCFBblock(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CFB (block) method of encryption }
    PROCEDURE DecryptCFBblock(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CFB (block) method of decryption }
    PROCEDURE EncryptOFB(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt
      size bytes of data using the OFB method of encryption }
    PROCEDURE DecryptOFB(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the OFB method of decryption }
    PROCEDURE EncryptCTR(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CTR method of encryption }
    PROCEDURE DecryptCTR(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CTR method of decryption }
  END;

{******************************************************************************}
    { Base type definition for 128 bit block ciphers }
TYPE
  TDCP_blockcipher128= CLASS(TDCP_blockcipher)
  PRIVATE
    IV, CV: ARRAY[0..15] OF BYTE;

    PROCEDURE IncCounter;
  PUBLIC
    CLASS FUNCTION GetBlockSize: INTEGER; override;
      { Get the block size of the cipher (in bits) }

    PROCEDURE Reset; override;
      { Reset any stored chaining information }
    PROCEDURE Burn; override;
      { Clear all stored key information and chaining information }
    PROCEDURE SetIV(CONST Value); override;
      { Sets the IV to Value and performs a reset }
    PROCEDURE GetIV(VAR Value); override;
      { Returns the current chaining information, not the actual IV }
    PROCEDURE Init(CONST Key; Size: longword; InitVector: POINTER); override;
      { Do key setup based on the data in Key, size is in bits }

    PROCEDURE EncryptCBC(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CBC method of encryption }
    PROCEDURE DecryptCBC(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CBC method of decryption }
    PROCEDURE EncryptCFB8bit(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CFB (8 bit) method of encryption }
    PROCEDURE DecryptCFB8bit(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CFB (8 bit) method of decryption }
    PROCEDURE EncryptCFBblock(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CFB (block) method of encryption }
    PROCEDURE DecryptCFBblock(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CFB (block) method of decryption }
    PROCEDURE EncryptOFB(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the OFB method of encryption }
    PROCEDURE DecryptOFB(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the OFB method of decryption }
    PROCEDURE EncryptCTR(CONST Indata; VAR Outdata; Size: longword); override;
      { Encrypt size bytes of data using the CTR method of encryption }
    PROCEDURE DecryptCTR(CONST Indata; VAR Outdata; Size: longword); override;
      { Decrypt size bytes of data using the CTR method of decryption }
  END;

IMPLEMENTATION


{** TDCP_blockcipher64 ********************************************************}

PROCEDURE TDCP_blockcipher64.IncCounter;
VAR
  i: INTEGER;
BEGIN
  Inc(CV[7]);
  i:= 7;
  WHILE (i> 0) AND (CV[i] = 0) DO
  BEGIN
    Inc(CV[i-1]);
    Dec(i);
  END;
END;

CLASS FUNCTION TDCP_blockcipher64.GetBlockSize: INTEGER;
BEGIN
  Result:= 64;
END;

PROCEDURE TDCP_blockcipher64.Init(CONST Key; Size: longword; InitVector: POINTER);
BEGIN
  inherited Init(Key,Size,InitVector);
  InitKey(Key,Size);
  IF InitVector= NIL THEN
  BEGIN
    FillChar(IV,8,{$IFDEF DCP1COMPAT}$FF{$ELSE}0{$ENDIF});
    EncryptECB(IV,IV);
    Reset;
  END
  ELSE
  BEGIN
    Move(InitVector^,IV,8);
    Reset;
  END;
END;

PROCEDURE TDCP_blockcipher64.SetIV(CONST Value);
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  Move(Value,IV,8);
  Reset;
END;

PROCEDURE TDCP_blockcipher64.GetIV(VAR Value);
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  Move(CV,Value,8);
END;

PROCEDURE TDCP_blockcipher64.Reset;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized')
  ELSE
    Move(IV,CV,8);
END;

PROCEDURE TDCP_blockcipher64.Burn;
BEGIN
  FillChar(IV,8,$FF);
  FillChar(CV,8,$FF);
  inherited Burn;
END;

PROCEDURE TDCP_blockcipher64.EncryptCBC(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    Move(p1^,p2^,8);
    XorBlock(p2^,CV,8);
    EncryptECB(p2^,p2^);
    Move(p2^,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.DecryptCBC(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
  Temp: ARRAY[0..7] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    Move(p1^,p2^,8);
    Move(p1^,Temp,8);
    DecryptECB(p2^,p2^);
    XorBlock(p2^,CV,8);
    Move(Temp,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.EncryptCFB8bit(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  Temp: ARRAY[0..7] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO Size DO
  BEGIN
    EncryptECB(CV,Temp);
    p2^:= p1^ xor Temp[0];
    Move(CV[1],CV[0],8-1);
    CV[7]:= p2^;
    Inc(p1);
    Inc(p2);
  END;
END;

PROCEDURE TDCP_blockcipher64.DecryptCFB8bit(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  TempByte: BYTE;
  Temp: ARRAY[0..7] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO Size DO
  BEGIN
    TempByte:= p1^;
    EncryptECB(CV,Temp);
    p2^:= p1^ xor Temp[0];
    Move(CV[1],CV[0],8-1);
    CV[7]:= TempByte;
    Inc(p1);
    Inc(p2);
  END;
END;

PROCEDURE TDCP_blockcipher64.EncryptCFBblock(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,8);
    XorBlock(p2^,CV,8);
    Move(p2^,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.DecryptCFBblock(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  Temp: ARRAY[0..7] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    Move(p1^,Temp,8);
    EncryptECB(CV,CV);
    Move(p1^,p2^,8);
    XorBlock(p2^,CV,8);
    Move(Temp,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.EncryptOFB(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,8);
    XorBlock(p2^,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.DecryptOFB(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,8);
    XorBlock(p2^,CV,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,CV,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.EncryptCTR(CONST Indata; VAR Outdata; Size: longword);
VAR
  temp: ARRAY[0..7] OF BYTE;
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,8);
    XorBlock(p2^,temp,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,temp,Size mod 8);
  END;
END;

PROCEDURE TDCP_blockcipher64.DecryptCTR(CONST Indata; VAR Outdata; Size: longword);
VAR
  temp: ARRAY[0..7] OF BYTE;
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 8) DO
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,8);
    XorBlock(p2^,temp,8);
    p1:= POINTER(longword(p1) + 8);
    p2:= POINTER(longword(p2) + 8);
  END;
  IF (Size mod 8)<> 0 THEN
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,Size mod 8);
    XorBlock(p2^,temp,Size mod 8);
  END;
END;

{** TDCP_blockcipher128 ********************************************************}

PROCEDURE TDCP_blockcipher128.IncCounter;
VAR
  i: INTEGER;
BEGIN
  Inc(CV[15]);
  i:= 15;
  WHILE (i> 0) AND (CV[i] = 0) DO
  BEGIN
    Inc(CV[i-1]);
    Dec(i);
  END;
END;

CLASS FUNCTION TDCP_blockcipher128.GetBlockSize: INTEGER;
BEGIN
  Result:= 128;
END;

PROCEDURE TDCP_blockcipher128.Init(CONST Key; Size: longword; InitVector: POINTER);
BEGIN
  inherited Init(Key,Size,InitVector);
  InitKey(Key,Size);
  IF InitVector= NIL THEN
  BEGIN
    FillChar(IV,16,{$IFDEF DCP1COMPAT}$FF{$ELSE}0{$ENDIF});
    EncryptECB(IV,IV);
    Reset;
  END
  ELSE
  BEGIN
    Move(InitVector^,IV,16);
    Reset;
  END;
END;

PROCEDURE TDCP_blockcipher128.SetIV(CONST Value);
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  Move(Value,IV,16);
  Reset;
END;

PROCEDURE TDCP_blockcipher128.GetIV(VAR Value);
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  Move(CV,Value,16);
END;

PROCEDURE TDCP_blockcipher128.Reset;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized')
  ELSE
    Move(IV,CV,16);
END;

PROCEDURE TDCP_blockcipher128.Burn;
BEGIN
  FillChar(IV,16,$FF);
  FillChar(CV,16,$FF);
  inherited Burn;
END;

PROCEDURE TDCP_blockcipher128.EncryptCBC(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    Move(p1^,p2^,16);
    XorBlock(p2^,CV,16);
    EncryptECB(p2^,p2^);
    Move(p2^,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.DecryptCBC(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
  Temp: ARRAY[0..15] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    Move(p1^,p2^,16);
    Move(p1^,Temp,16);
    DecryptECB(p2^,p2^);
    XorBlock(p2^,CV,16);
    Move(Temp,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.EncryptCFB8bit(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  Temp: ARRAY[0..15] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO Size DO
  BEGIN
    EncryptECB(CV,Temp);
    p2^:= p1^ xor Temp[0];
    Move(CV[1],CV[0],15);
    CV[15]:= p2^;
    Inc(p1);
    Inc(p2);
  END;
END;

PROCEDURE TDCP_blockcipher128.DecryptCFB8bit(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  TempByte: BYTE;
  Temp: ARRAY[0..15] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO Size DO
  BEGIN
    TempByte:= p1^;
    EncryptECB(CV,Temp);
    p2^:= p1^ xor Temp[0];
    Move(CV[1],CV[0],15);
    CV[15]:= TempByte;
    Inc(p1);
    Inc(p2);
  END;
END;

PROCEDURE TDCP_blockcipher128.EncryptCFBblock(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,16);
    XorBlock(p2^,CV,16);
    Move(p2^,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.DecryptCFBblock(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: Pbyte;
  Temp: ARRAY[0..15] OF BYTE;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    Move(p1^,Temp,16);
    EncryptECB(CV,CV);
    Move(p1^,p2^,16);
    XorBlock(p2^,CV,16);
    Move(Temp,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.EncryptOFB(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,16);
    XorBlock(p2^,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.DecryptOFB(CONST Indata; VAR Outdata; Size: longword);
VAR
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,16);
    XorBlock(p2^,CV,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,CV);
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,CV,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.EncryptCTR(CONST Indata; VAR Outdata; Size: longword);
VAR
  temp: ARRAY[0..15] OF BYTE;
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,16);
    XorBlock(p2^,temp,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,temp,Size mod 16);
  END;
END;

PROCEDURE TDCP_blockcipher128.DecryptCTR(CONST Indata; VAR Outdata; Size: longword);
VAR
  temp: ARRAY[0..15] OF BYTE;
  i: longword;
  p1, p2: POINTER;
BEGIN
  IF NOT fInitialized THEN
    raise EDCP_blockcipher.Create('Cipher not initialized');
  p1:= @Indata;
  p2:= @Outdata;
  FOR i:= 1 TO (Size div 16) DO
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,16);
    XorBlock(p2^,temp,16);
    p1:= POINTER(longword(p1) + 16);
    p2:= POINTER(longword(p2) + 16);
  END;
  IF (Size mod 16)<> 0 THEN
  BEGIN
    EncryptECB(CV,temp);
    IncCounter;
    Move(p1^,p2^,Size mod 16);
    XorBlock(p2^,temp,Size mod 16);
  END;
END;

END.
